% Matlab code example for Algorithm 2 presented in "Sparse Fast DCT for 
% Vectors with One-block Support" by Sina Bittens and Gerlind Plonka.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ x, supp_struct, mu, m ] = DCT_shortsupp( dct_x,  epsilon, varargin )
%DCT_SHORTSUPP Reconstructs the vector x with short support from its DCT-II
% dct_x using DFT_reflected.

% Since computing the full vector yhat from dct_x would be inefficient, we 
% use a modified version of DFT_reflected in this algorithm which computes 
% every required entry of yhat at the point in the algorithm where it is 
% actually used.

% The length N of x has to be a power of 2. Write it as N=2^(J-1). The 
% vector y that is reconstructed by DFT_reflected has length 2N=2^J.
N=length(dct_x);
if abs(log2(N)-round(log2(N)))<eps 
    J=log2(N)+1;
    % Check whether a lower bound on the block length of y is known. 
    if nargin==3
        b=varargin{:};
        if b>J
            error('2^b can be at most length(y).')
        end
    else
        % If no bound is known, start with a vector of length 2^0=1.
        b=0;
    end
    % Initialize the vector for yhat=fft(y).
    yhat=zeros(2*N,1);
    % Compute the smallest periodized vector. If b=0, then y=yhat(1).
    % Compute the required 2^b entries of yhat.
    yhat((2^(J-b)*(0:floor((N-1)/2^(J-b))))+1)=sqrt(2*N)*(exp(2*pi*1i*2^(J-b)*(0:floor((N-1)/2^(J-b)))/(4*N))).'.*dct_x(2^(J-b)*(0:floor((N-1)/2^(J-b)))+1);
    % epsilon_N(l)=1/sqrt(2) for l=0.
    yhat(1)=yhat(1)*sqrt(2);
    % yhat(N+1) is already 0.
    yhat((2^(J-b)*(ceil((N+1)/2^(J-b)):2^b-1))+1)=-sqrt(2*N)*(exp(2*pi*1i*2^(J-b)*(ceil((N+1)/2^(J-b)):2^b-1)/(4*N))).'.*dct_x(2*N-2^(J-b)*(ceil((N+1)/2^(J-b)):2^b-1)+1);
    % epsilon_N(2N-2^(J-b)*l)=1 for all l in ceil((N+1)/2^(J-b)):2^b-1.
    % Compute y^(b) from the 2^b entries of yhat.
    y=ifft(yhat((2^(J-b)*(0:2^b-1))+1));
    % If b==0, then y has a one-block support. Otherwise detect its support
    % structure.
    if b==0
        supp_struct=1;
        mu=0;
        m=1;
    else
        [supp_struct, mu, m]=find_supp_0(y); 
    end
    % In each step recover y^(j+1) from y^(j) and additional entries of
    % yhat using the appropriate method for the given support structure.
    for j=b:J-1
        %% Recovery Step for a Periodization with One-block Support
        if supp_struct==1
            % If m>2^(j-1), y^(j) may not have a uniquely defined first
            % support index and y^(j) does not have a short support we can
            % benefit from. Hence, recover y^(j+1) via an IFFT of length
            % 2^j.
            if m>2^(j-1)
                % Compute the required 2^j entries of yhat.
                yhat(2^(J-j-1)*(2*(0:floor(1/2*((N-1)/2^(J-j-1)-1)))+1)+1)=sqrt(2*N)*(exp(2*pi*1i*2^(J-j-1)*(2*(0:floor(1/2*((N-1)/2^(J-j-1)-1)))+1)/(4*N))).'.*dct_x(2^(J-j-1)*(2*(0:floor(1/2*((N-1)/2^(J-j-1)-1)))+1)+1);
                yhat(2^(J-j-1)*(2*(ceil(1/2*((N+1)/2^(J-j-1)-1)):2^j-1)+1)+1)=-sqrt(2*N)*(exp(2*pi*1i*2^(J-j-1)*(2*(ceil(1/2*((N+1)/2^(J-j-1)-1)):2^j-1)+1)/(4*N))).'.*dct_x(2*N-2^(J-j-1)*(2*(ceil(1/2*((N+1)/2^(J-j-1)-1)):2^j-1)+1)+1);
                % epsilon_N(l)=1 for all indices l corresponding to these 
                % entries of yhat. yhat(N) is not one of them.                
                data=yhat(2^(J-j-1)*(2*(0:2^j-1)+1)+1);
                a=exp(pi*1i*(0:2^j-1)/2^j).'.*ifft(data);
                % Since y is a non-negative real vector, we can disregard
                % the imaginary part. Further, we apply a thresholding
                % procedure in order to be able to better determine the
                % support structure of y^(j+1).
                y_0=real(0.5*(y+a));
                aux=find(y_0<=epsilon);
                y_0(aux)=zeros(size(aux));
                % Use that all periodizations of y are symmetric.
                y_1=y_0(end:-1:1);
                y=[y_0;y_1]; 
                % Find the support structure of y^(j+1).
                [supp_struct, mu, m, mu2, m2]=find_supp_1(y,j,J,mu,m,epsilon);
            % If m<=2^(j-1), the first support index of y^(j) is uniquely
            % determined and we can benefit from the fact that the support 
            % of y^(j) is short.    
            else
                L=ceil(log2(m));
                % Restrict to 2^L equations and variables.
                % Compute the 2^L required entries of yhat.
                yhat(2^(J-L)*(0:floor((N-1-2^(J-j-1))/2^(J-L)))+2^(J-j-1)+1)=sqrt(2*N)*(exp(2*pi*1i*(2^(J-L)*(0:floor((N-1-2^(J-j-1))/2^(J-L)))+2^(J-j-1))/(4*N))).'.*dct_x(2^(J-L)*(0:floor((N-1-2^(J-j-1))/2^(J-L)))+2^(J-j-1)+1);
                yhat(2^(J-L)*(ceil((N+1-2^(J-j-1))/2^(J-L)):2^L-1)+2^(J-j-1)+1)=-sqrt(2*N)*(exp(2*pi*1i*(2^(J-L)*(ceil((N+1-2^(J-j-1))/2^(J-L)):2^L-1)+2^(J-j-1))/(4*N))).'.*dct_x(2*N-(2^(J-L)*(ceil((N+1-2^(J-j-1))/2^(J-L)):2^L-1)+2^(J-j-1))+1);
                % epsilon_N(l)=1 for all indices l corresponding to these 
                % entries of yhat. yhat(N) is not one of them.                
                data=yhat(2^(J-L)*(0:2^L-1)+2^(J-j-1)+1);
                z=y(mod(mu+(0:(2^L-1)),2^j)+1);
                % If mu>2^j-2^L, the modulo operation in the first matrix
                % has to be taken into account.
                if mu>2^j-2^L
                   a=[exp(pi*1i*(mu+(0:2^j-mu-1))/2^j) -exp(pi*1i*(mu+(2^j-mu:2^L-1))/2^j)].'.*ifft(exp(2*pi*1i*mu*(0:2^L-1)/2^L).'.*data);
                else
                   a=exp(pi*1i*(mu+(0:2^L-1))/2^j).'.*ifft(exp(2*pi*1i*mu*(0:2^L-1)/2^L).'.*data);
                end               
                % Since y is a non-negative real vector, we can disregard
                % the imaginary part. Further, we apply a thresholding
                % procedure in order to be able to better determine the
                % support structure of y^(j+1).
                z=real(0.5*(z+a));
                aux=find(z<=epsilon);
                z(aux)=zeros(size(aux));
                y_0=zeros(2^j,1);
                y_1=zeros(2^j,1);
                y_0(mod(mu+(0:(2^L-1)),2^j)+1)=z(1:2^L);
                % Use that all periodizations of y are symmetric.
                y_1(2^j-1-mod(mu+(0:2^L-1),2^j)+1)=z(1:2^L);
                y=[y_0;y_1];
                % Find the support structure of y^(j+1).
                [supp_struct, mu, m, mu2, m2]=find_supp_2(y,j,J,mu,m,epsilon);
            end
        %% Recovery Step for a Periodization with Two-Block Support
        elseif supp_struct==2
            % If y^(j) has a two-block support, then y^(j+1) has a 
            % two-block support as well and we only need to find the first 
            % support index of the first block of y^(j) in y^(j+1).
            % Find a non-zero oddly indexed entry of fft(y^(j+1)) by only
            % considering O(m) entries.
            % Compute the required 2*m entries of yhat.
            d=min(floor((N-1-2^(J-j-1))/2^(J-j)),2*m-1);
            dd=min(ceil((N-1-2^(J-j-1))/2^(J-j)),2*m-1);
            yhat(2^(J-j)*(0:d)+2^(J-j-1)+1)=sqrt(2*N)*(exp(2*pi*1i*(2^(J-j)*(0:d)+2^(J-j-1))/(4*N))).'.*dct_x(2^(J-j)*(0:d)+2^(J-j-1)+1);
            if dd<2*m-1
                yhat(2^(J-j)*(dd:2*m-1)+2^(J-j-1)+1)=-sqrt(2*N)*(exp(2*pi*1i*(2^(J-j)*(dd:2*m-1)+2^(J-j-1))/(4*N))).'.*dct_x(2*N-(2^(J-j)*(dd:2*m-1)+2^(J-j-1))+1);
            end
            [~,k]=max(abs(yhat(2^(J-j-1)*(2*(0:2*m-1)+1)+1)));
            k=k-1;
            yhat_odd=yhat(2^(J-j-1)*(2*k+1)+1);           
            % Compute the 2k+1-st entry of the DFT of the vector u that has
            % y^(j) as its periodization and the same first support index 
            % as y^(j).
            uhat=sum(y((mu:mu+m-1)+1).*(exp(-pi*1i*(2*k+1)/2^j*(mu:mu+m-1))).')+sum(y(2^j-1-(mu:mu+m-1)+1).*(exp(pi*1i*(2*k+1)/2^j*(mu+1:mu+m))).');
            % If uhat and yhat_odd are equal, then the first support index
            % lambda of the first block of y^(j) in y^(j+1) is mu, so y^(j)
            % and y^(j+1) have the same first support index.
            if abs(uhat-yhat_odd)<abs(uhat+yhat_odd)
                lambda=mu;
            % Otherwise the first support index of that block in y^(j+1) is
            % 2^j+mu.
            else
                lambda=2^j+mu;
            end
            y_0=zeros(2^(j+1),1);
            y_0((lambda:lambda+m-1)+1)=y(mod((lambda:lambda+m-1),2^j)+1);
            % Use that all periodizations of y are symmetric.
            y_0((2^(j+1)-m-lambda:2^(j+1)-1-lambda)+1)=y(mod((2^(j+1)-m-lambda:2^(j+1)-1-lambda),2^j)+1);
            y=y_0;
            % Find the support structure of y^(j).
            [supp_struct, mu, m]=find_supp_3(j,mu,m,lambda);
        end
    end  
    x=y(1:N);
    % If y has a two-block support, then the support length of x is the 
    % length m of one block in y and the first support index in x is the 
    % first support index mu in y.
    % If y has a one-block support, then the support length of x is m/2, 
    % half the length of the support length of y.
    if supp_struct==1
        m=m/2;
        % If the first support index in y is greater than N-1, the first
        % support index in x is the reflection of the last support index in
        % y.
        if mu>=N 
            mu=2*N-m-mu; 
        end
    % If y has a two-block support with blocks of possibly different
    % lengths, then the first support index in x is the first support index
    % mu of the first block in y and the support length is (m+m2)/2.
    elseif supp_struct==3
        m=(m+m2)/2;
    end
else
    disp('The length of dct_x has to be a power of 2.')
end

end

