% Required for the Matlab code example for Algorithm 2 presented in "Sparse 
% Fast DCT for Vectors with One-block Support" by Sina Bittens and Gerlind
% Plonka.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [  ] = TEST_DCT_shortsupp( tests,m,N,epsilon,variant )
%TEST_DCT_SHORTSUPP Computes the average runtime and reconstruction error
%of DCT_shortsupp and MATLAB's idct for an input vector of length N with 
%short or one-block support.

% If variant==0, the average runtimes of DCT_shortsupp and idct are 
% computed for randomly generated noiseless data. If variant==2, the 
% average reconstruction errors for randomly generated noisy input vectors
% (uniform noise with SNR values [0 10 ... 50]) are computed for 
% DFT_reflected and idct.

% epsilon is the threshold for cutting off noise required by DFT_reflected 
% in order to find the support structure of the reconstructed vectors.

% For DFT_reflected, and hence for DCT_shortsupp, the length of the input 
% vector has to be a power of 2.
if abs(log2(N)-round(log2(N)))<eps
    J=log2(N)+1;
    M=length(m);
    % Initialize the matrix for the average runtimes.
    time=zeros(M,4);
    time(:,1)=m;
    % Define the SNR values for noisy signals.
    SNR=zeros(6,1);
    for l=1:length(SNR)
        SNR(l)=(l-1)*10;
    end
    % Initialize the matrix of reconstruction errors for exact data.
    error_ex=zeros(M,2);    
    for j=1:M
        % Initialize the matrix for the average runtimes for exact data.
        time_noise=zeros(length(SNR),3);
        time_noise(:,1)=SNR;
        % Initialize the matrix for the average noisy reconstruction 
        % errors.
        error=zeros(length(SNR),3);
        error(:,1)=SNR;
        % Initialize the matrix for the rate of correct recovery.
        correct_recov=zeros(length(SNR),3);
        correct_recov(:,1)=SNR;
        % Define the cutoff thresholds for noisy data, optimized for
        % N=2^20.
        epsilon_noise(6)=0.05;
        epsilon_noise(5)=0.15;
        epsilon_noise(4)=0.3;
        epsilon_noise(3)=1;
        epsilon_noise(2)=1.8;
        epsilon_noise(1)=2.5;
        for t=1:tests
            if mod(t,10)==1
                t
            end
            % For each test create a random first support index mu for the
            % support block of x, which has to be a vector of length 
            % N=2^(J-1) with short support.
            mu=randi([0 N-1],1);
            x=zeros(N,1);
            % Create the m(j) support indices.
            supp=[0:mu+m(j)-1-N,mu:min(mu+m(j)-1,N-1)]; 
            % Generate a random number between 0 and floor((m(j)-2)/2) and
            % create that many zeros inside the support, but not at the 
            % first or last support index.
            suppzeros=randi(floor((m(j)-2)/2)+1)-1; 
            % Select suppzeros elements randomly from 2,...,m-1 and remove
            % the respective entries from supp.
            supp(randperm(m(j)-2,suppzeros)+1)=[];
            % Create the m(j)-suppzeros random non-negative entries of the 
            % support.
            x(supp+1)=rand(m(j)-suppzeros,1)*10;
            % Create the input data.
            dct_x=dct(x);
            
            %% Standard IDCT-II Algorithm
            % Measure the runtime of idct.
            tic;
            S_idct=idct(dct_x);
            time(j,3)=time(j,3)+toc;
            
            %% Numerical Experiments for Exact Input Data
            if variant==0
                %% DCT_shortsupp
                % Measure the runtime of DCT_shortsupp.
                tic;
                [S_shortsupp, ~, ~, ~] =DCT_shortsupp(dct_x,epsilon);
                time(j,2)=time(j,2)+toc;
                
                % Compute the reconstruction errors for exact data for the 
                % two methods, given by the norm of the difference between 
                % the original vector x and the returned vectors
                % S_shortsupp and S_idct.
                error_ex(j,1)=error_ex(j,1)+norm(S_shortsupp-x)/N;
                error_ex(j,2)=error_ex(j,2)+norm(S_idct-x)/N;  
            
            %% Numerical Experiments for Noisy Input Data    
            elseif variant==1
                for n=1:length(SNR)
                    % Create random uniformly distributed noise for the 
                    % input vector dct_x.
                    noise=rand(N,1)-0.5;
                    % Scale the noise according to the current SNR value
                    % SNR(n).
                    noise=norm(dct_x)*10^(-SNR(n)/20)/norm(noise)*noise;
                    % Create noisy input data by addition.
                    dct_x_noise=dct_x+noise;
                    
                    %% Standard IDCT for Noisy Input Data
                    % Measure the runtime of idct for noisy input data.
                    tic;
                    S_idct_noise=idct(dct_x_noise);
                    time_noise(n,3)=time_noise(n,3)+toc;
                    
                    %% DCT_shortsupp for Noisy Input Data
                    % Measure the runtime of DCT_shortsupp for noisy input
                    % data.
                    tic;
                    [S_shortsupp_noise, ~, mu_short_n, m_short_n]=DCT_shortsupp(dct_x_noise,epsilon_noise(n));
                    time_noise(n,2)=time_noise(n,2)+toc;
                    % Check whether the support of x was found correctly.
                    % First examine whether the support of x is contained
                    % in the returned support of S_shortsupp_noise, and, if
                    % necessary, increase the counter correct_recov(n,2).
                    if mu_short_n+m_short_n<=N-1
                        if (mu>=mu_short_n && mu+m(j)-1<=mu_short_n+m_short_n-1) 
                            correct_recov(n,2)=correct_recov(n,2)+1;
                            % We also introduce a second counter
                            % correct_recov(n,3) which is only increased if
                            % the returned support length of
                            % S_shortsupp_noise is at most three times the
                            % length of the support length of x.
                            if (m_short_n <= 3* m(j) && m_short_n>=m(j))
                                correct_recov(n,3)=correct_recov(n,3)+1;
                            end
                        end
                    else
                        % If 0, N-1 are contained in the support of
                        % S_shortsupp_noise, the block has to be considered
                        % modulo N.
                        if ((mu>=mu_short_n || mu<=mu_short_n+m_short_n-1-N) && (mu+m(j)-1>=mu_short_n || mu+m(j)-1<=mu_short_n+m_short_n-1-N))
                            correct_recov(n,2)=correct_recov(n,2)+1;
                            if  (m_short_n <= 3* m(j) && m_short_n>=m(j))
                                correct_recov(n,3)=correct_recov(n,3)+1;
                            end
                        end
                    end
                    % Compute the reconstruction errors for noisy data for 
                    % the two methods, given by the norm of the difference 
                    % between the original vector x and the returned
                    % vectors S_shortsupp_noise and S_idct_noise.
                    error(n,2)=error(n,2)+norm(S_shortsupp_noise-x)/N;
                    error(n,3)=error(n,3)+norm(S_idct_noise-x)/N;
                end
            end
        end
        % Compute the average runtimes and the average reconstruction 
        % errors.
        time(j,1)=m(j);
        time(j,4)=N;
        time_noise(:,2:3)=time_noise(:,2:3)/tests
        error(:,2:3)=error(:,2:3)/tests
        correct_recov(:,2:3)=correct_recov(:,2:3)/tests
        % Save the matrix of reconstruction error values in a separate file
        %for each support length m(j).
        if variant ==1
            filename = ['error_DCT_shortsupp' num2str(m(j)) '.dat'];
            dlmwrite(filename, error(:,:), 'delimiter', '\t');
            % Save the matrix of correct recovery rates in a file.
            filename = ['correctrecov_DCT' num2str(m(j)) '.dat'];
            dlmwrite(filename,correct_recov,'delimiter','\t');
            % Save the matrix of noisy runtime values in a file.
            filename = ['runtimeDCT_short_noise_' num2str(m(j)) '.dat'];
            dlmwrite(filename,time_noise,'delimiter','\t');
        end
    end
    time(:,2:3)=time(:,2:3)/tests
    error_ex=error_ex/tests
    if variant ==0
        % Save the matrix of runtime values in a file.
        dlmwrite('runtimeDCT_short.dat',time,'delimiter','\t');
        % Save the matrix of exact errors in a file.
        dlmwrite('error_exact_DCT.dat',error_ex,'delimiter','\t');
    end
end

end

    


