% Required for the Matlab code example for Algorithm 1 presented in "Sparse 
% Fast DCT for Vectors with One-block Support" by Sina Bittens and Gerlind
% Plonka.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ supp_struct, mu1, m1,mu2,m2 ] = find_supp_1( y,j,J,mu,m,epsilon)
%FIND_SUPP_1 Finds the support structure of y=y^(j+1) if y^(j) has a 
%one-block support with length m>2^(j-1).
% supp_struct==1 if y has a one-block support, supp_struct==2 if y has a
% two-block support and supp_struct==3 if it has a two-block support with 
% possibly different block lenghts (only possible for j=J-1).

% Since we do not know whether supp_struct==3, we set the parameters of the
% second, possibly short or longer, block to 0.
mu2=0;
m2=0;
% supp_struct==3 is only possible for j==J-1.
if j<J-1 
    % If 2^(j-1)-1,2^(j-1) are contained in S^(j) and supp(y^(j)) is not 
    % the whole interval 0,...,2^j-1, then y^(j+1) has a two-block support.
    if (mu+m-1<=2^j-1 && ((2^(j-1)-1>=mu) && (2^(j-1)-1<=mu+m-1) && (2^(j-1)>=mu) && (2^(j-1)<=mu+m-1) && m~=2^j ))  ||  (mu+m-1>2^j-1 && ((2^(j-1)-1>=mu  || 2^(j-1)-1<=mod(mu+m-1,2^j)) && (2^(j-1)>=mu  || 2^(j-1)<=mod(mu+m-1,2^j)) && m~=2^j))
        supp_struct=2;
        y_0=y(1:2^j);
        % Find the first support index of the first block.
        T=find(y_0>epsilon)-1;
        mu1=T(1);
        m1=T(end)-T(1)+1;
    % Otherwise y^(j+1) has a one-block support.
    else
        supp_struct=1;
        y_0=y(1:2^j);
        % Find the significant entries in the first half of y^(j+1).
        T_0=find(y_0>epsilon)-1;
        K=length(T_0);
        if K==0
            % If there are none, then the noisy data must have cancelled
            % out the significant entries. Since we cannot find the support
            % structure, we set the first support index to 0 and the length
            % to 2^(j+1), in the hope that a direct DFT of length 2^(j+1)
            % in the next step helps recovering the data.
            mu1=0;
            m1=2^(j+1);
        else
            % Otherwise find out whether the support block is centered
            % around 0 and 2^(j+1)-1 or around 2^j-1 and 2^j. 
            T=[T_0;2^(j+1)-1-(T_0(K:-1:1))];
            e_0=T(K+1)-T(K);
            e_1=mod(T(1)-T(2*K),2^(j+1));
            % If more than half of the entries of y^(j+1) are significant,
            % the first support index is not uniquely defined anymore.
            % Hence, we set the first support index to 0 and the block
            % length to 2^(j+1), in the hope that a direct DFT of length
            % 2^(j+1) in the next step yields a support that is shorter
            % than half of the current vector length.
            if T(2*K)-T(1)+1>=2^j
                supp_struct=1;
                mu1=0;
                m1=2^(j+1);
            else
                if e_0<e_1
                    % If the distance between the last significant entry in
                    % the first half of y^(j+1) and the first significant
                    % entry in the second half is greater than the periodic
                    % distance between the first significant entry in the
                    % first half and the last significant entry in the
                    % second half, the support is centered around 2^j-1 and
                    % 2^j.
                    m1=T(2*K)-T(1)+1;
                    mu1=T(1);
                elseif e_1<e_0
                    % If e_1<e_0, the support is centered around 0 and
                    % 2^(j+1).
                    m1=2^(j+1)-T(K+1)+T(K)+1;
                    mu1=T(K+1);
                else
                    % If the distances are the same, both y^(j) and y^(j+1)
                    % have full support and we set the first support index
                    % to 0.
                    m1=2^(j+1);
                    mu1=0;
                end
            end
        end
    end
else
    % If j=J-1, we have the same cases as before, but it might also happen
    % that y has a two-block support with two blocks of different lengths.
    % If 2^(j-1)-1,2^(j-1) are contained in S^(j) and the support of y^(j) 
    % is not the whole interval 0,...,2^j-1, then y^(j+1) has a two-block 
    % support.
    % 2^(j-1)-1,2^(j-1) in S^(j)
    if (mu+m-1<=2^j-1 && ((2^(j-1)-1>=mu) && (2^(j-1)-1<=mu+m-1) && (2^(j-1)>=mu) && (2^(j-1)<=mu+m-1) && m~=2^j ))  ||  (mu+m-1>2^j-1 && ((2^(j-1)-1>=mu  || 2^(j-1)-1<=mod(mu+m-1,2^j)) && (2^(j-1)>=mu  || 2^(j-1)<=mod(mu+m-1,2^j)) && m~=2^j))
        supp_struct=2;
        y_0=y(1:2^j);
        % Find the first support index of the first block.
        T=find(y_0>epsilon)-1;
        mu1=T(1);
        m1=T(end)-T(1)+1;
    % If 0, 2^j-1 are contained in S^(j), then y^(j+1) has a two-block
    % support with possibly different block lengths, and one of the blocks
    % might be empty.
    elseif (mu+m-1<=2^j-1 && ((0>=mu) && (0<=mu+m-1) && (2^j-1>=mu) && (2^j-1<=mu+m-1) ))  ||  (mu+m-1>2^j-1 && ((0>=mu  || 0<=mod(mu+m-1,2^j)) && (2^j-1>=mu  || 2^j-1<=mod(mu+m-1,2^j)) ))
        if m~=2^j
            % If the support of y^(j) is not the whole interval
            % 0,...,2^j-1, we can uniquely determine the support structure.
            y_0=y((mu:2^j-1)+1);
            y_1=y((2^j+mu:2^(j+1)-1)+1);
            % Find the relevant entries of y^(j+1).
            T_0=find(y_0>epsilon)-1+mu;
            T_1=find(y_1>epsilon)-1+2^j+mu;
            % Not both index sets T_0 and T_1 can be empty.
            if isempty(T_0)==1
                % If T_0 is empty, then y^(j+1) has only a one-block 
                % support and 0,2^J-1 are contained in it.
                supp_struct=1;
                mu1=T_1(1);
                m1=2*(2^(j+1)-mu1);
            elseif isempty(T_1)==1
                % If T_1 is empty, then y^(j+1) has only a one-block 
                % support and 2^(J-1)-1, 2^(J-1) are contained in it.
                supp_struct=1;
                mu1=T_0(1);
                m1=2*(2^j-mu1);            
            else
                % If neither set is empty, then y^(j+1) has a two-block 
                % support with two blocks that might have different
                % lengths. This can only happen if 0, 2^(J-1)-1 are 
                % contained in the support of x.                
                supp_struct=3;            
                mu1=T_0(1);
                m1=2*(2^j-mu1);
                mu2=T_1(1);
                m2=2*(2^J-mu2);
            end
        else
            % If the support of y^(j) is the whole interval, the support is
            % not always uniquely to recover, as there is no unique first
            % support index and block length of y^(j) that one can use for
            % y^(j+1).
            y_0=y(1:m/2);
            y_1=y(m/2+1:2^j);
            % Find the relevant entries of the first half of y^(j+1).
            T_0=find(y_0>epsilon)-1+m/2;
            T_1=find(y_1>epsilon)-1;
            % Not both index sets T_0 and T_1 can be empty.
            if isempty(T_0)==1
                % If T_0 is empty, then y^(j+1) has only a one-block 
                % support and 2^(J-1)-1 and 2^(J-1) are contained in it.
                supp_struct=1;
                mu1=T_1(1);
                m1=2*(2^(J-1)-T_1(1));
            elseif isempty(T_1)==1
                % If T_1 is empty, then y^(j+1) has only a one-block 
                % support and 0, 2^J-1 are contained in it.
                supp_struct=1;
                mu1=2^J-1-T_0(end);
                m1=2*(T_0(end)+1);            
            else
                % If neither set is empty, there are three possibilities
                % for the support of y^(J), and we do not know the
                % correct one a priori. Either
                % - S^(J)={T_0(1),...,2^(J-1)-1,2^(J-1),...,2^J-1-T_0(1)},
                % - S^(J)={2^J-1-T_1(end),...,2^J-1,0,T_1(end),
                % - y^(J) has a two-block support with two blocks of
                % possibly different lengths, where neither the position of
                % the blocks nor their length is known.
                % The first and second case can only happen if the first
                % support index of x is 0 or the last support index of x is
                % 2^J-1, respectively. 
                supp_struct=3;            
                mu1=T_0(1);
                m1=2*(2^(J-1)-mu1);
                mu2=2^J-1-T_1(end);
                m2=2*(T_1(end)+1);
                fprintf(' Support is either %d,...,%d,%d,...,%d or %d,...,%d,0,...,%d or a two-block support with unknwon block lengths and positions.\n', T_0(1),2^(J-1)-1,2^(J-1),2^J-1-T_0(1),2^J-1-T_1(end),2^J-1,T_1(end))
            end            
        end
    end
end

end

