import numpy as np
import math
from scipy.fftpack import ifft


def DCT_shortsupp(dct_x, epsilon, b=0):
    """
    Reconstructs the vector x with one-block support of length from its DCT-II dct_x using Algorithm 1 in "Sparse fast
    DCT for vectors with one-block support" by Sina Bittens and Gerlind Plonka.
    :param dct_x: DCT-II of the vector x with one-block support we aim to recover
    :param epsilon: noise threshold
    :param b: lower bound on the block length m of y = (x, (Jx)^T)^T: 2^(b-1) <= m
    :return x:
    :return supp_struct: support structure of y
             'one' if y has a one-block support
             'two' if y has a two-block support
             'two_diff' if y has a two-block support with possibly different lengths, only possible for j = J-1
    :return mu: first support index of x
    :return m: support length of x
    """

    # DCT_shortsupp uses a modified version of DFT_reflected, as computing the full vector yhat from dct_x is
    # inefficient. Instead, every required entry of yhat is computed at the point in the algorithm where it is used for
    # the first time.

    dct_x = dct_x.ravel()
    N = dct_x.shape[0]
    # The length N of dct_x has to be a power of 2. Write it as N=2^(J-1).
    if abs(math.log(N, 2)-round(math.log(N, 2))) < np.spacing(1):
        J = round(math.log(N, 2)) + 1
        # The optional lower bound on the block length of y can be at most
        if b > J:
            print('Error: b = {} > J = {}.'.format(b, J))
            return
        # Initialize the vector for yhat = fft(y).
        yhat = np.zeros((2*N), dtype=complex)
        # Compute 2^b entries of yhat corresponding to the DFT of the smallest periodized vector. If b == 0, then
        # y = yhat[0].
        yhat[2**(J - b) * np.arange(0, math.floor((N-1) / 2**(J - b)) + 1)] \
            = math.sqrt(2 * N) * (np.exp(2 * np.pi * 1j * 2**(J-b) *
                                         np.arange(0, math.floor((N - 1) / 2**(J - b)) + 1) / (4 * N))) * \
              dct_x[2**(J - b) * np.arange(0, math.floor((N-1) / 2**(J - b)) + 1)]
        # epsilon_N(l) = 1 / sqrt(2) for l = 0.
        yhat[0] = yhat[0] * math.sqrt(2)
        # yhat[N] is already 0.
        yhat[2**(J - b) * np.arange(math.ceil((N + 1) / 2**(J - b)), 2**b)] \
            = -math.sqrt(2 * N) * (np.exp(2 * np.pi * 1j * 2**(J-b) *
                                          np.arange(math.ceil((N + 1) / 2**(J - b)), 2**b) / (4 * N))) *\
              dct_x[2*N - 2**(J-b) * np.arange(math.ceil((N+1) / 2**(J-b)), 2**b)]
        # epsilon_N(2*N-2^(J-b)*l) = 1 for all l in ceil((N+1)/2^(J-b)),...,2^b-1.
        # Compute y^(b) from the 2^b entries of yhat calculated above.
        y = ifft(yhat[2**(J - b) * np.arange(0, 2**b)])

        # If b == 0, then y has a one-block support. Otherwise detect its support structure.
        if b == 0:
            supp_struct = 'one'
            mu = 0
            m = 1
        else:
            supp_struct, mu, m = find_supp_0(y, epsilon)

        # In each step recover y^(j+1) from y^(j) and additional entries of yhat using the appropriate method for the
        # given supprt structure.
        for j in range(b, J):
            # Recovery step for a periodization y^(j) with one-block support.
            if supp_struct == 'one':
                # If m > 2^(j-1), y^(j) may not have a uniquely defined first support index and y^(j) does not have a
                # short support we can benefit from. Hence, recover y^(j+1) via an IFFT of length 2^j.
                if m > 2**(j-1):
                    # Compute the required 2^j entries of yhat.
                    yhat[2**(J-j-1) * (2 * np.arange(0, math.floor(1 / 2 * ((N - 1) / 2**(J - j - 1) - 1)) + 1) + 1)] \
                        = math.sqrt(2 * N) * (np.exp(2 * np.pi * 1j * 2**(J - j - 1) *
                                                     (2 * np.arange(0, math.floor(1 / 2 * ((N - 1) / 2**(J - j - 1) - 1)
                                                                                  ) + 1) + 1) / (4 * N))) * \
                        dct_x[2**(J-j-1) * (2 * np.arange(0, math.floor(1 / 2 * ((N - 1) / 2**(J - j - 1) - 1)) + 1)
                                            + 1)]
                    yhat[2**(J-j-1) * (2 * np.arange(math.ceil(1 / 2 * ((N + 1) / 2**(J - j - 1) - 1)), 2**j) + 1)] \
                        = -math.sqrt(2 * N) * (np.exp(2 * np.pi * 1j * 2**(J - j - 1) *
                                                      (2 * np.arange(math.ceil(1 / 2 * ((N + 1) / 2**(J - j - 1) - 1)
                                                                               ), 2**j) + 1) / (4 * N))) \
                        * dct_x[2 * N - 2**(J-j-1) * (2 * np.arange(math.ceil(1 / 2 * ((N + 1) / 2**(J - j - 1) - 1)
                                                                              ), 2**j) + 1)]
                    # epsilon_N(l) = 1 for all indices l corresponding to these entries of yhat. y[N] is not one of
                    # them.
                    data = yhat[2**(J-j-1) * (2 * np.arange(0, 2**j) + 1)]
                    a = np.exp(np.pi * 1j * np.arange(0, 2**j) / 2**j) * ifft(data)
                    # Since y is a non-negative real vector, we can disregard the imaginary part. Further, we apply a
                    # thresholding procedure in order to be able to better determine the support structure of y^(j+1).
                    y_0 = np.real(0.5 * (y + a))
                    aux = np.nonzero(abs(y_0) <= epsilon)[0]
                    y_0[aux] = np.zeros(aux.shape)
                    # Use that all periodizations of y are symmetric.
                    y_1 = np.flip(y_0, 0)
                    y = np.concatenate((y_0, y_1))
                    # Find the support structure of y^(j+1).
                    supp_info = find_supp_1(y, j, J, mu, m, epsilon)
                    supp_struct, mu, m = supp_info[0: 3]
                    if supp_struct == 'two_diff':
                        mu_2, m_2 = supp_info[3:]

                # If m <= 2**j, then the first support index of y^(j) is uniquely determined and we can benefit from
                # the fact that the support of y^(j) is short.
                else:
                    L = math.ceil(math.log(m, 2))
                    # Restrict to 2^L equations and variables.
                    # Compute the 2^L required entries of yhat.
                    yhat[2**(J-L) * np.arange(0, math.floor((N - 1 - 2**(J - j - 1)) / 2**(J - L)) + 1)
                         + 2**(J - j - 1)] \
                        = math.sqrt(2 * N) * (np.exp(2 * np.pi * 1j * (2**(J - L) *
                                                                       np.arange(0, math.floor((N - 1 - 2**(J - j - 1))
                                                                                               / 2**(J - L)) + 1)
                                                                       + 2**(J - j - 1)) / (4 * N))) \
                        * dct_x[2**(J - L) * np.arange(0, math.floor((N - 1 - 2**(J - j - 1)) / 2**(J - L)) + 1)
                                + 2**(J - j - 1)]
                    yhat[2**(J-L) * np.arange(math.ceil((N + 1 - 2**(J-j-1)) / 2**(J - L)), 2**L) + 2**(J-j-1)] \
                        = -math.sqrt(2 * N) * (np.exp(2 * np.pi * 1j * (2**(J - L) *
                                                                        np.arange(math.ceil((N + 1 - 2**(J - j - 1)) /
                                                                                            2**(J - L)), 2**L)
                                                                        + 2**(J - j - 1)) / (4 * N)))\
                        * dct_x[2 * N - (2**(J - L) * np.arange(math.ceil((N + 1 - 2**(J - j - 1)) / 2**(J - L)), 2**L)
                                         + 2**(J - j - 1))]
                    # epsilon_N(l) = 1 for all indices l corresponding to these entries. yhat[N] is not one of them.
                    data = yhat[2**(J - L) * np.arange(0, 2**L) + 2**(J - j - 1)]
                    z = y[(mu + np.arange(0, 2**L)) % 2**j]
                    # If mu>2^j-2^L, the modulo operation in the first matrix has to be taken into account.
                    if mu > 2**j - 2**L:
                        a = np.concatenate((np.exp(np.pi * 1j * (mu + np.arange(0, 2**j - mu)) / 2**j),
                                            -np.exp(np.pi * 1j * (mu + np.arange(2**j - mu, 2**L)) / 2**j))) \
                            * ifft(np.exp(2 * np.pi * 1j * mu * np.arange(0, 2**L) / 2**L) * data)
                    else:
                        a = np.exp(np.pi * 1j * (mu + np.arange(0, 2**L)) / 2**j) \
                            * ifft(np.exp(2 * np.pi * 1j * mu * np.arange(0, 2**L) / 2**L) * data)
                    # Since y is a non-negative real vector, we can disregard its imaginary part. Further, we apply a
                    # thresholding procedure in order to be able to better determine the support structure of y^(j+1).
                    z = np.real(0.5 * (z + a))
                    aux = np.nonzero(abs(z) <= epsilon)[0]
                    z[aux] = np.zeros(aux.shape)
                    y_0 = np.zeros(2**j)
                    y_1 = np.zeros(2**j)
                    y_0[(mu + np.arange(0, 2**L)) % 2**j] = z[0: 2**L]
                    # Use that all periodizations of y are symmetric.
                    y_1[2**j - 1 - (mu + np.arange(0, 2**L)) % 2**j] = z[0: 2**L]
                    y = np.concatenate((y_0, y_1))
                    # Find the support structure of y^(j+1).
                    supp_info = find_supp_2(y, j, J, mu, m, epsilon)
                    supp_struct, mu, m = supp_info[0: 3]
                    if supp_struct == 'two_diff':
                        mu_2, m_2 = supp_info[3:]

            # Recovery step for a periodization y^(j+1) with two-block support.
            elif supp_struct == 'two':
                # If y^(j) has a two-block support, then y^(j+1) has a two-block support as well and we only need to
                # find the first support index of the first block of y^(j) in y^(j+1).
                # Find an oddly indexed non-zero entry of fft(y^(j+1)) by only considering O(m) entries.
                # Compute the required 2*m entries of yhat.
                d = min([math.floor((N - 1 - 2**(J-j-1)) / 2**(J-j)), 2 * m - 1])
                yhat[2**(J-j) * np.arange(0, d + 1) + 2**(J - j - 1)] \
                    = math.sqrt(2 * N) * (np.exp(2 * np.pi * 1j * (2**(J-j) * np.arange(0, d + 1) + 2**(J-j-1)) /
                                                 (4 * N))) \
                    * dct_x[2**(J-j) * np.arange(0, d + 1) + 2**(J-j-1)]

                if d+1 < 2 * m - 1:
                    yhat[2**(J-j) * np.arange(d+1, 2 * m) + 2**(J-j-1)] \
                        = -math.sqrt(2 * N) * (np.exp(2 * np.pi * 1j * (2**(J-j) * np.arange(d+1, 2 * m) + 2**(J-j-1)) /
                                                      (4 * N))) \
                        * dct_x[2 * N - (2**(J-j) * np.arange(d+1, 2 * m) + 2**(J-j-1))]

                k = np.argmax(abs(yhat[2**(J - j - 1) * (2 * np.arange(0, 2 * m) + 1)]))
                yhat_odd = yhat[2**(J-j-1) * (2 * k + 1)]
                # Compute the 2k+1-st entry of the DFT of the vector u that has y^(j) as its periodization and the same
                # first support index as y^(j).
                uhat = np.sum(y[mu: mu + m] * (np.exp(-np.pi * 1j * (2 * k + 1) / 2**j * np.arange(mu, mu + m)))) \
                     + np.sum(y[2**j-1-np.arange(mu, mu + m)] * (np.exp(np.pi * 1j * (2 * k + 1) / 2**j
                                                                 * np.arange(mu, mu + m))))
                # If uhat and yhat_odd are equal, then the first support index l of the first block of y^(j) in y^(j+1)
                # is mu, so y^(j) and y^(j+1) have the same first support index.
                if abs(uhat - yhat_odd) < abs(uhat + yhat_odd):
                    l = mu
                # Otherwise the first support index of that block in y^(j+1) is 2^j+mu.
                else:
                    l = 2**j + mu
                y_0 = np.zeros(2**(j + 1))
                y_0[l: l + m] = y[np.arange(l, l + m) % 2**j]
                # Use that all periodizations of y are symmetric.
                y_0[2**(j + 1) - m - l: 2**(j + 1) - l] = y[np.arange(2**(j + 1) - m - l, 2**(j + 1) - l) % 2**j]
                y = y_0
                # Find the support structure of y^(j).
                supp_struct, mu, m = find_supp_3(j, mu, m, l)

        x = y[0: N]
        # If y has a two-block support, then the support length of x is the length m of one block in y and the first
        # support index of x is the first support index mu of y. If y has a one-block support, then the support length
        # of x is m/2.
        if supp_struct == 'one':
            m = (m / 2)
            # If the first support index of y is greater than N-1, the first support index of x is the reflection of the
            # last support index of y.
            if mu > N:
                mu = (2 * N - m - mu)
        # If y has a two-block support with blocks of possibly different lengths, then the first support index of x is
        # the first support index mu of the first block in y and the support length is (m+m_2)/2.
        elif supp_struct == 'two_diff':
            m = ((m + m_2) / 2)
        # print(x)
        return x, supp_struct, mu, m

    else:
        print('The length of dct_x has to be a power of 2.')


def find_supp_0(y, epsilon):
    """
    Detects the support structure of y if nothing about its support is known.
    :param y: periodization of a vector with reflected block support
    :param epsilon: noise threshold
    :return: support structure:
             'one' if y has a one-block support
             'two' if y has a two-block support
    :return: mu: first support index
    :return: m: support length
    """
    y = y.ravel()
    N = y.shape[0]
    # Since y is symmetric, it suffices to examine the first half of it.
    y_0 = y[0: round(N / 2)]
    T = np.nonzero(abs(y_0) > epsilon)[0]

    if T[0] == 0 and T[-1] != round(N / 2) - 1:
        # y has a one-block support of length m < N.
        supp_struct = 'one'
        mu = N - 1 - T[-1]
        m = 2 * (T[-1] + 1)
    elif T[-1] == round(N / 2) - 1 and T[0] != 0:
        # y has a one-block support of length m<N.
        supp_struct = 'one'
        mu = T[0]
        m = 2 * (T[-1] - T[0] + 1)
    elif T[-1] == round(N / 2) - 1 and T[0] == 0:
        # y has a one-block support of length m = N.
        supp_struct = 'one'
        mu = 0
        m = N
    else:
        # y has a two-block support of block length m < N/".
        supp_struct = 'two'
        mu = T[0]
        m = T[-1] - T[0] + 1
    return supp_struct, mu, m


def find_supp_1(y, j, J, mu, m, epsilon):
    """
    Finds the support structure of y = y^(j+1) if y^(j) has a one-block support of length m > 2^(j-1).
    :param y:
    :param j:
    :param J: full vector has length 2^J
    :param mu: first support index of y^(j)
    :param m: support length of y^(j)
    :param epsilon: noise threshold
    :return supp_struct: support structure of y
             'one' if y has a one-block support
             'two' if y has a two-block support
             'two_diff' if y has a two-block support with possibly different lengths, only possible for j = J-1
    :return mu_1: first support index of the first block of y^(j+1)
    :return m_1: length of the first block of y^(j+1)
    :return mu_2: first support index of the second block of y^(j+1)
    :return m_2: length of the second block of y^(j+1)
    """

    if j < J - 1:
        # If 2^(j-1)-1, 2^(j-1) are contained in S^(j) and supp((y^(j)) is not the whole interval 0,...,2^j-1, then
        # y has a two-block support.
        if (mu + m - 1 <= 2**j - 1 and ((2**(j-1) - 1 >= mu) and (2**(j-1) - 1 <= mu + m - 1) and (2**(j-1) >= mu) and
            (2**(j-1) <= mu + m - 1) and m != 2**j)) \
                or (mu + m - 1 > 2**j - 1 and ((2**(j-1) - 1 >= mu or 2**(j-1) - 1 <= (mu + m - 1) % 2**j) and
                                               (2**(j-1) >= mu or 2**(j-1) <= (mu + m - 1) % 2**j) and m != 2**j)):
            supp_struct = 'two'
            y_0 = y[0: 2**j]
            # Find the first support index of the first block.
            T = np.nonzero(abs(y_0) > epsilon)[0]
            mu_1 = T[0]
            m_1 = T[-1] - T[0] + 1
            return supp_struct, mu_1, m_1
        # Otherwise y has a one-block support.
        else:
            supp_struct = 'one'
            y_0 = y[0: 2**j]
            # Find the significant entries in the first half of y.
            T_0 = np.nonzero(abs(y_0) > epsilon)[0]
            K = T_0.shape[0]
            if K == 0:
                # If there are none, then the noisy data has cancelled out the significant entries. Set the first
                # support index to 0 and the length to 2^(j+1).
                mu_1 = 0
                m_1 = 2**(j+1)
                return supp_struct, mu_1, m_1
            else:
                # Otherwise detect whether the support block is centered around 0 and 2^(j+1)-1 or around 2^j-1 and 2^j.
                T = np.concatenate((T_0, 2**(j+1) - 1 - np.flip(T_0, 0)))
                # print('here')

                e_0 = T[K] - T[K - 1]
                e_1 = (T[0] - T[2*K - 1]) % 2**(j+1)
                # If more than half of the entries of y are significant, the first support index is not uniquely defined
                # anymore. We set the first support index to 0 and the block length to 2^(j+1).
                # if T[2 * K - 1] - T[0] + 1 > 2**j:
                if 2 * K > 2 ** j:
                    supp_struct = 'one'
                    mu_1 = 0
                    m_1 = 2**(j+1)
                    return supp_struct, mu_1, m_1
                else:
                    if e_0 < e_1:
                        # If the distance between the last significant entry in the first half of y and the first
                        # significant entry in the second half is greater than the periodic distance between the first
                        # significant entry in the first half and the last significant entry in the second half, the
                        # support is centered around 2^j-1 and 2^j.
                        m_1 = T[2*K - 1] - T[0] + 1
                        mu_1 = T[0]
                    elif e_1 < e_0:
                        # If e_1 < e_0, the support is centered around 0 and 2^(j+1).
                        m_1 = 2**(j+1) - T[K] + T[K - 1] + 1
                        mu_1 = T[K]
                    else:
                        # If the distances are the same, both y^(j) and y^(j+1) have full support and we set the first
                        # support index to 0.
                        m_1 = 2**(j+1)
                        mu_1 = 0
                    return supp_struct, mu_1, m_1

    else:
        # If j = J-1, we have the same cases as before, but it might also happen that y has a two-block support with
        # two blocks of different lengths.
        # If 2^(j-1)-1, 2^(j-1) are contained in S^(j) and the support of y^(j) is not the whole interval 0,...,2^j-1,
        # then y^(j+1) has a two-block support.
        # 2^(j-1)-1, 2^(j-1) in S^(j).
        if (mu + m - 1 <= 2**j - 1 and ((2**(j - 1) - 1 >= mu) and (2**(j - 1) - 1 <= mu + m - 1) and (2**(j - 1) >= mu)
            and (2**(j - 1) <= mu + m - 1) and m != 2**j)) or (mu + m - 1 > 2**j - 1
                                                               and ((2**(j - 1) - 1 >= mu
                                                                     or 2**(j - 1) - 1 <= (mu + m - 1) % 2**j)
                                                                    and (2**(j - 1) >= mu
                                                                         or 2**(j - 1) <= (mu + m - 1) % 2**j)
                                                                    and m != 2**j)):
            supp_struct = 'two'
            y_0 = y[0: 2**j]
            # Find the first support index of the first block.
            T = np.nonzero(abs(y_0) > epsilon)[0]
            mu_1 = T[0]
            m_1 = T[-1] - T[0] + 1
            return supp_struct, mu_1, m_1
        # If 0, 2^j-1 are contained in S^(j), then y has a two-block support with possibly different block lengths, and
        # one of the blocks might be empty.
        elif (mu + m - 1 <= 2**j - 1 and ((0 >= mu) and (0 <= mu + m - 1) and (2**j - 1 >= mu) and
                                          (2**j - 1 <= mu + m - 1))) or (mu + m - 1 > 2**j - 1 and
                                                                         ((0 >= mu or 0 <= (mu + m - 1) % 2**j)
                                                                          and (2**j - 1 >= mu or
                                                                               2**j - 1 <= (mu + m - 1) % 2**j))):
            if m != 2**j:
                # If the support of y^(j) is not the whole interval 0,...,2^j-1, we can uniquely determine the support
                # structure.
                y_0 = y[mu: 2**j]
                y_1 = y[2**j + mu: 2**(j+1)]
                # Find the significant large entries of y.
                T_0 = np.nonzero(abs(y_0) > epsilon)[0] + mu
                T_1 = np.nonzero(abs(y_1) > epsilon)[0] + 2**j + mu
                # The index sets T_0 and T_1 cannot both be empty.
                if T_0.shape[0] == 0:
                    # If T_0 is empty, then y has only a one-block support and 0, 2^J-1 are contained in it.
                    supp_struct = 'one'
                    mu_1 = T_1[0]
                    m_1 = 2 * (2**(j+1) - mu_1)
                    return supp_struct, mu_1, m_1
                elif T_1.shape[0] == 0:
                    # If T_1 is empty, then y has only a one-block support and 2^(J-1)-1, 2^(J-1) are contained in it.
                    supp_struct = 'one'
                    mu_1 = T_0[0]
                    m_1 = 2 * (2**j - mu_1)
                    return supp_struct, mu_1, m_1
                else:
                    # If neither set is empty, then y has a two-block support with two blocks that might have different
                    # lengths. This can only happen if 0, 2^(J-1)-1 are contained in the support of x.
                    supp_struct = 'two_diff'
                    mu_1 = T_0[0]
                    m_1 = 2 * (2**j - mu_1)
                    mu_2 = T_1[0]
                    m_2 = 2 * (2**J - mu_2)
                    return supp_struct, mu_1, m_1, mu_2, m_2
            else:
                # If the support of y^(j) is the whole interval, the support cannot always be uniquely recovered, as
                # there are no unique first support index and block length of y^(j) that one can use for y^(j+1).
                y_0 = y[0: round(m / 2)]
                y_1 = y[round(m / 2): 2**j]
                # Find the significantly large entries of the first half of y.
                T_0 = np.nonzero(abs(y_0) > epsilon)[0]
                T_1 = np.nonzero(abs(y_1) > epsilon)[0] + round(m / 2)
                # The index sets T_0 and T_1 cannot both be empty.
                if T_0.shape[0] == 0:
                    # If T_0 is empty, then y has only a one-support and 2^(J-1)-1, 2^(J-1) are contained in it.
                    supp_struct = 'one'
                    mu_1 = T_1[0]
                    m_1 = 2 * (2**(J-1) - T_1[0])
                    return supp_struct, mu_1, m_1
                elif T_1.shape[0] == 0:
                    # If T_1 is empty, then y has only a one-support and 0, 2^J-1 are contained in it.
                    supp_struct = 'one'
                    mu_1 = 2**J - 1 - T_0[-1]
                    m_1 = 2 * (T_0[-1] + 1)
                    return supp_struct, mu_1, m_1
                else:
                    # If neither set is empty, there are three possibilitis for the support of y, and we do not know the
                    # correct one a priori. Either
                    # - S^(J) = {T_0[0],...,2^(J-1)-1,2^(J-1),...,2^J-1-T_0[0]},
                    # - S^(J) = {2^J-1-T_1[-1],...,2^J-1,0,...,T_1[-1],
                    # - y^(J) has a two-block support with two blocks of possibly different lengths, where neither the
                    # position of the blocks nor their length is known.
                    # The first and second case can only happen if the first support index of x is 0 or the last support
                    # index of x is 2^(J-1)-1, respectively.
                    supp_struct = 'two_diff'
                    mu_1 = T_0[0]
                    m_1 = 2 * (2**(J-1) - mu_1)
                    mu_2 = 2**J - 1 - T_1[-1]
                    m_2 = 2 * (T_1[-1] + 1)
                    print('Support is either {},...,{},{},...,{} or {},...,{},0,...,{} or a two-block support with '
                          'unknown block lengths and positions.'.format(T_0[0], 2**(J - 1) - 1, 2**(J - 1),
                                                                        2**J - 1 - T_0[0], 2**J - 1 - T_1[-1], 2**J - 1,
                                                                        T_1[-1]))
                    return supp_struct, mu_1, m_1, mu_2, m_2


def find_supp_2(y, j, J, mu, m, epsilon):
    """
    Finds the support structure of y = y^(j+1) if y^(j) has a one-block support of length m <= 2^(j-1).
    :param y: vector whose support we want to determine
    :param j: y has length 2^(j+1)
    :param J: full vector has length 2^J
    :param mu: first support index of y^(j)
    :param m: support length of y^(j)
    :param epsilon: noise threshold
    :return supp_struct: support structure of y
             'one' if y has a one-block support
             'two' if y has a two-block support
             'two_diff' if y has a two-block support with possibly different lengths, only possible for j = J-1
    :return mu_1: first support index of the first block of y^(j+1)
    :return m_1: length of the first block of y^(j+1)
    :return mu_2: first support index of the second block of y^(j+1)
    :return m_2: length of the second block of y^(j+1)
    """
    if j < J-1:
        # If 2^(j-1)-1, 2^(j-1) are contained in S^(j), then y^(j+1) has a two-block support.
        if (mu + m - 1 <= 2**j - 1 and ((2**(j-1) - 1 >= mu) and (2**(j-1) - 1 <= mu + m - 1) and (2**(j-1) >= mu)
           and (2**(j-1) <= mu + m - 1))) or (mu + m - 1 > 2**j - 1 and ((2**(j-1) - 1 >= mu
                                                                          or 2**(j-1) - 1 <= (mu + m - 1) % 2**j)
                                                                         and (2**(j-1) >= mu
                                                                              or 2**(j-1) <= (mu + m - 1) % 2**j))):
            supp_struct = 'two'
            y_0 = y[mu: 2**j - mu]
            # Find the first support index of the first block.
            T = np.nonzero(abs(y_0) > epsilon)[0] + mu
            mu_1 = T[0]
            m_1 = T[-1] - T[0] + 1
            return supp_struct, mu_1, m_1

        # If 0, 2^j-1 are contained in S^(j), then y^(j+1) has a one-block support.
        elif (mu + m - 1 <= 2**j - 1 and ((0 >= mu) and (0 <= mu + m - 1) and (2**j - 1 >= mu)
                                          and (2**j - 1 <= mu + m - 1))) \
                or (mu + m - 1 > 2**j - 1 and ((0 >= mu or 0 <= (mu + m - 1) % 2**j) and (2**j - 1 >= mu
                                                                                          or 2**j - 1 <=
                                                                                          (mu + m - 1) % 2**j))):
            supp_struct = 'one'
            # Find the first support index of the block.
            e_0 = y[mu: mu + m].sum()
            e_1 = y[(np.arange(mu, mu + m) + 2**j + 1) % 2**(j+1)].sum()
            if e_0 > e_1:
                mu_1 = mu
            elif e_0 < e_1:
                mu_1 = mu + 2**j
            m_1 = m
            return supp_struct, mu_1, m_1

    else:
        # If j=J-1, we have the same cases as before, but it might also happen that y has a two-block support with two
        # blocks of different lengths.
        # If 2^(j-1)-1, 2^(j-1) are contained in S^(j), then y has a two-block support.
        if (mu + m - 1 <= 2**j - 1 and (
                (2**(j - 1) - 1 >= mu) and (2**(j - 1) - 1 <= mu + m - 1) and (2**(j - 1) >= mu) and (
                2**(j - 1) <= mu + m - 1))) or (mu + m - 1 > 2**j - 1 and (
                (2**(j - 1) - 1 >= mu or 2**(j - 1) - 1 <= (mu + m - 1) % 2**j) and (
                2**(j - 1) >= mu or 2**(j - 1) <= (mu + m - 1) % 2**j))):
            supp_struct = 'two'
            y_0 = y[mu: 2**j - mu]
            # Find the first support index of the first block.
            T = np.nonzero(abs(y_0) > epsilon)[0] + mu
            mu_1 = T[0]
            m_1 = T[-1] - T[0] + 1
            return supp_struct, mu_1, m_1
        # If 0, 2^j-1 are contained in S^(j), then y has a two-block support with possibly different block lengths, and
        # one of the blocks might be empty.
        elif (mu + m - 1 <= 2**j - 1 and (
                    (0 >= mu) and (0 <= mu + m - 1) and (2**j - 1 >= mu) and (2**j - 1 <= mu + m - 1))) or (
                    mu + m - 1 > 2**j - 1 and ((0 >= mu or 0 <= (mu + m - 1) % 2**j) and (
                        2**j - 1 >= mu or 2**j - 1 <= (mu + m - 1) % 2**j))):
            y_0 = y[mu: 2**j]
            y_1 = y[2**j + mu:]
            # Find the significant entries of y.
            T_0 = np.nonzero(abs(y_0) > epsilon)[0] + mu
            T_1 = np.nonzero(abs(y_1) > epsilon)[0] + 2**j + mu
            # Not both index sets T_0 and T_1 can be empty.
            if T_0.shape[0] == 0:
                # If T_0 is empty, then y has only a one-block support, and 0, 2^J-1 are contained in it.
                supp_struct = 'one'
                mu_1 = T_1[0]
                m_1 = 2 * (2**(j + 1) - mu_1)
                return supp_struct, mu_1, m_1
            elif T_1.shape[0] == 0:
                # If T_1 is empty, then y has only a one-block support, and 2^(J-1)-1, 2^(J-1) are contained in it.
                supp_struct = 'one'
                mu_1 = T_0[0]
                m_1 = 2 * (2**j - mu_1)
                return supp_struct, mu_1, m_1
            else:
                # If neither set is empty, then y has a two-block support with two blocks that might have different
                # lengths. This can only happen if 0, 2^(J-1)-1 are contained in the support of x.
                supp_struct = 'two_diff'
                mu_1 = T_0[0]
                m_1 = 2 * (2**j - mu_1)
                mu_2 = T_1[0]
                m_2 = 2 * (2**J - mu_2)
                return supp_struct, mu_1, m_1, mu_2, m_2


def find_supp_3(j, mu, m, l):
    """
    Finds the support structure of y^(j+1) if y^(j) has a two-block support.
    :param j: y^(j+1) has length 2^(j+1)
    :param mu: first support index of y^(j)
    :param m: block length of y^(j)
    :param l: parameter depending on the support structure of y^(j+1)
    :return supp_struct: support structure of y
         'one' if y has a one-block support
         'two' if y has a two-block support
         'two_diff' if y has a two-block support with possibly different lengths, only possible for j = J-1
    :return mu_1: first support index of the first block of y^(j+1)
    :return m_1: length of the first block of y^(j+1)
    """

    # The first support index of the first block only depends on the previously computed values l, mu, m and j. y^(j)
    # and y^(j+1) both have a two-block support with the same block length.
    supp_struct = 'two'
    if l == mu:
        mu_1 = mu
        return supp_struct, mu_1, m
    elif l == 2**j + mu:
        mu_1 = 2**j - m - mu
        return supp_struct, mu_1, m
